from __future__ import annotations

import json
import logging
import random
import string
import threading
import traceback
import uuid
from typing import Any, Callable, TypeVar

import autogen  # type: ignore
from autogen import Agent
from autogen.logger.base_logger import BaseLogger  # type: ignore
from autogen.logger.logger_utils import get_current_ts, to_dict  # type: ignore
from openai import AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletion

from okareo import Okareo
from okareo.model_under_test import ModelUnderTest

logger = logging.getLogger(__name__)

F = TypeVar("F", bound=Callable[..., Any])


def agent_obj_to_name(agent: Agent) -> str:  # type: ignore
    return agent.name if isinstance(agent, Agent) else agent  # type: ignore


def safe_serialize(obj: Any) -> str:
    def default(o: Any) -> str:
        if hasattr(o, "to_json"):
            return str(o.to_json())
        elif callable(o):
            return f"<function {o.__name__}>"
        else:
            return f"<<non-serializable: {type(o).__qualname__}>>"

    return json.dumps(obj, default=default)


class OkareoLogger(BaseLogger):  # type: ignore
    # For more info on BaseLogger class, see Autogen docs.
    # https://microsoft.github.io/autogen/0.2/docs/reference/logger/base_logger
    def __init__(
        self,
        config: dict[str, Any],
    ) -> None:
        assert "api_key" in config, "api_key is required in the config"
        api_key = config["api_key"]
        base_path = config.get("base_path", None)

        if base_path and len(base_path) > 0:
            self.okareo = Okareo(api_key, base_path=base_path)
        else:
            self.okareo = Okareo(api_key)

        self.context_token = str(config.get("context_token", ""))

        random_suffix = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=5)
        )
        self.tags = config.get("tags", [])
        self.group_name = config.get("group_name", f"autogen-chat-{random_suffix}")
        self.group = self.okareo.create_group(
            name=self.group_name,
            tags=self.tags,
            source={"log_source": "autogen"},
        )
        self.agent_names: list[str] = []
        self.registered_models: dict[str, ModelUnderTest] = {}
        self.cur_model = None

        # set default logging options
        self._log_chat_completion = config.get("log_chat_completion", True)
        self._log_function_use = config.get("log_function_use", True)
        self._log_new_agent = config.get("log_new_agent", True)
        self._log_event = config.get("log_event", True)
        self._log_new_wrapper = config.get("log_new_wrapper", False)
        self._log_new_client = config.get("log_new_client", False)
        self._log_types = [
            "chat_completion",
            "function_use",
            "new_agent",
            "event",
            "new_wrapper",
            "new_client",
        ]

        # set verbosity
        self.verbose = config.get("verbose", False)

    def register_new_model(self, agent_name: str, tags: list[str]) -> None:
        """Register an agent as a new model in Okareo, and add the model to the group."""
        new_model = self.okareo.register_model(
            name=agent_name,
            tags=tags,
        )
        self.okareo.add_model_to_group(self.group, new_model)
        self.registered_models[agent_name] = new_model
        if self.verbose:
            print(f"[Okareo] New agent registered under model: {agent_name}")

    def start(self) -> str:
        """Start the logger and return the session_id."""
        self.session_id = (
            self.context_token if len(self.context_token) > 0 else str(uuid.uuid4())
        )
        try:
            if self.verbose:
                print(
                    f"[Okareo] Started new logging session with context_token: {self.session_id}"
                )
                print(
                    f"[Okareo] Logging data points under group_name '{self.group_name}'."
                )
                print("[Okareo] Logging the following events:")
            for log_type in self._log_types:
                if getattr(self, f"_log_{log_type}"):
                    print(f"[Okareo] - {log_type}")
        except Exception as e:
            print(f"[Okareo] Failed to start logging: {e}")
        return self.session_id

    def log_chat_completion(
        self,
        invocation_id: uuid.UUID,
        client_id: int,
        wrapper_id: int,
        source: Any,  # TODO: str | Agent,
        request: dict[str, float | str | list[dict[str, str]]],
        response: str | ChatCompletion,
        is_cached: int,
        cost: float,
        start_time: str,
    ) -> None:
        """
        Log a chat completion to Okareo. From the Autogen docs:

        "In AutoGen, chat completions are somewhat complicated because they are handled by the autogen.oai.OpenAIWrapper class. One invocation to create can lead to multiple underlying OpenAI calls, depending on the llm_config list used, and any errors or retries."
        """
        if self._log_chat_completion:
            thread_id = threading.get_ident()
            source_name = None
            if isinstance(source, str):
                source_name = getattr(source, "name", "unknown")
            else:
                source_name = source.name
            try:
                response_dict = json.loads(
                    safe_serialize(
                        response if isinstance(response, str) else response.model_dump()
                    )
                )
                # register new model if need be
                if source_name not in self.registered_models:
                    self.register_new_model(source_name, self.tags)
                # parse the response to get the proper message
                self.registered_models[source_name].add_data_point(
                    input_obj=request,
                    input_datetime=start_time,
                    result_obj=json.dumps(
                        {
                            "log_type": "chat_completion",
                            "invocation_id": invocation_id,
                            "client_id": client_id,
                            "wrapper_id": wrapper_id,
                            "response": response_dict,
                            "is_cached": is_cached,
                            "cost": cost,
                            "thread_id": thread_id,
                            "source_name": source_name,
                        }
                    ),
                    result_datetime=get_current_ts(),
                    context_token=self.session_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
                if self.verbose:
                    print(f"[Okareo] Logged chat completion for agent '{source_name}'")
            except Exception:
                tb = traceback.format_exc()
                line_number = tb.splitlines()[-3].split(",")[1].strip().split()[1]
                print(
                    f"[Okareo] Failed to log chat completion (Line {line_number}): {tb}"
                )
        else:
            pass

    def log_new_agent(
        self, agent: Any, init_args: dict[str, Any]  # TODO: ConversableAgent,
    ) -> None:
        """
        Log the birth of a new agent to Okareo.
        """
        if self._log_new_agent:
            thread_id = threading.get_ident()
            agent_name = None
            if isinstance(agent, str):
                agent_name = getattr(agent, "name", "unknown")
            else:
                agent_name = agent.name
            try:
                # make init_args['chat_messages'] serializable
                if (
                    "chat_messages" in init_args
                    and init_args["chat_messages"] is not None
                ):
                    init_args["chat_messages"] = {
                        agent_obj_to_name(k): agent_obj_to_name(v)
                        for k, v in init_args["chat_messages"].items()
                    }
                # register new model if need be
                if agent_name not in self.registered_models:
                    self.register_new_model(agent_name, self.tags)
                # parse the response to get the proper message
                self.registered_models[agent_name].add_data_point(
                    result_obj=json.dumps(
                        {
                            "log_type": "new_agent",
                            "agent_id": id(agent),
                            "agent_name": (
                                agent.name
                                if hasattr(agent, "name") and agent.name is not None
                                else ""
                            ),
                            "wrapper_id": to_dict(
                                agent.client.wrapper_id
                                if hasattr(agent, "client") and agent.client is not None
                                else ""
                            ),
                            "session_id": self.session_id,
                            "current_time": get_current_ts(),
                            "agent_type": type(agent).__name__,
                            "args": safe_serialize(init_args),
                            "thread_id": thread_id,
                        }
                    ),
                    result_datetime=get_current_ts(),
                    context_token=self.session_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
                if self.verbose:
                    print(f"[Okareo] Logged new agent '{agent_name}'")
            except Exception:
                tb = traceback.format_exc()
                line_number = tb.splitlines()[-3].split(",")[1].strip().split()[1]
                print(f"[Okareo] Failed to log new agent (Line {line_number}): {tb}")
        else:
            pass

    def log_event(
        self, source: Any, name: str, **kwargs: dict[str, Any]  # TODO: str | Agent,
    ) -> None:
        """
        Log an event from an agent or a string source.
        """
        if self._log_event:
            # This takes an object o as input and returns a string. If the object o cannot be serialized, instead of raising an error,
            # it returns a string indicating that the object is non-serializable, along with its type's qualified name obtained using __qualname__.
            json_args = json.dumps(
                kwargs,
                default=lambda o: f"<<non-serializable: {type(o).__qualname__}>>",
            )
            thread_id = threading.get_ident()

            # register new model if need be
            source_name = str(source.name) if hasattr(source, "name") else source
            if source_name not in self.registered_models:
                self.register_new_model(source_name, self.tags)

            if isinstance(source, str):
                try:
                    # register new model if need be
                    self.registered_models[source_name].add_data_point(
                        result_obj=json.dumps(
                            {
                                "log_type": "event",
                                "source_id": id(source),
                                "source_name": source_name,
                                "event_name": name,
                                "json_state": json_args,
                                "timestamp": get_current_ts(),
                                "thread_id": thread_id,
                            }
                        ),
                        result_datetime=get_current_ts(),
                        context_token=self.session_id,
                        tags=self.tags,
                        group_id=self.group.get("id"),
                    )
                    if self.verbose:
                        print(
                            f"[Okareo] Logged event '{name}' for agent '{source_name}'"
                        )
                except Exception as e:
                    print(f"[Okareo] Failed to log event: {e}")
            else:
                try:
                    # parse the response to get the proper message
                    self.registered_models[source_name].add_data_point(
                        result_obj=json.dumps(
                            {
                                "log_type": "event",
                                "source_id": id(source),
                                "source_name": source_name,
                                "event_name": name,
                                "agent_module": source.__module__,
                                "agent_class": source.__class__.__name__,
                                "json_state": json_args,
                                "timestamp": get_current_ts(),
                                "thread_id": thread_id,
                            }
                        ),
                        result_datetime=get_current_ts(),
                        context_token=self.session_id,
                        tags=self.tags,
                        group_id=self.group.get("id"),
                    )
                except Exception as e:
                    print(f"[Okareo] Failed to log event: {e}")
        else:
            pass

    def log_new_wrapper(
        self,
        wrapper: Any,  # TODO: OpenAIWrapper
        init_args: dict[str, Any],  # TODO: dict[str, LLMConfig | list[LLMConfig]]
    ) -> None:
        """
        Log a new wrapper instance.
        """
        if self._log_new_wrapper:
            thread_id = threading.get_ident()
            try:
                # TODO: figure out if/where agent name occurs
                self.registered_model.add_data_point(
                    result_obj=json.dumps(
                        {
                            "log_type": "new_wrapper",
                            "wrapper_id": id(wrapper),
                            "session_id": self.session_id,
                            "json_state": safe_serialize(init_args),
                            "timestamp": get_current_ts(),
                            "thread_id": thread_id,
                        }
                    ),
                    result_datetime=get_current_ts(),
                    context_token=self.session_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
                if self.verbose:
                    print(f"[Okareo] Logged wrapper '{wrapper}''")
            except Exception:
                tb = traceback.format_exc()
                line_number = tb.splitlines()[-3].split(",")[1].strip().split()[1]
                print(f"[Okareo] Failed to log new wrapper (Line {line_number}): {tb}")
        else:
            pass

    def log_new_client(
        self,
        client: AzureOpenAI | OpenAI,
        wrapper: Any,  # TODO: OpenAIWrapper
        init_args: dict[str, Any],
    ) -> None:
        """
        Log a new client instance.
        """
        if self._log_new_client:
            thread_id = threading.get_ident()

            try:
                # TODO: figure out if/where agent name occurs
                self.registered_model.add_data_point(
                    result_obj=json.dumps(
                        {
                            "log_type": "new_client",
                            "client_id": id(client),
                            "wrapper_id": id(wrapper),
                            "session_id": self.session_id,
                            "class": type(client).__name__,
                            "json_state": json.dumps(init_args),
                            "timestamp": get_current_ts(),
                            "thread_id": thread_id,
                        }
                    ),
                    result_datetime=get_current_ts(),
                    context_token=self.session_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
                if self.verbose:
                    print(
                        f"[Okareo] Logged new_client '{client}' for wrapper '{wrapper}'"
                    )
            except Exception as e:
                print(f"[Okareo] Failed to log new client: {e}")
        else:
            pass

    def log_function_use(
        self,
        source: Any,  # TODO: str | Agent
        function: F,
        args: dict[str, Any],
        returns: Any,
    ) -> None:
        """
        Log a registered function (can be a tool) use from an agent or a string source.
        """
        if self._log_function_use:
            thread_id = threading.get_ident()
            if self.verbose:
                print(f"[Okareo] function use with:\n-> args={args}\nreturns={returns}")
            try:
                source_name = str(source.name) if hasattr(source, "name") else source
                if source_name not in self.registered_models:
                    self.register_new_model(source_name, self.tags)
                self.registered_models[source_name].add_data_point(
                    result_obj=json.dumps(
                        {
                            "log_type": "function_use",
                            "source_id": id(source),
                            "source_name": source_name,
                            "agent_module": source.__module__,
                            "agent_class": source.__class__.__name__,
                            "timestamp": get_current_ts(),
                            "thread_id": thread_id,
                            "input_args": safe_serialize(args),
                            "returns": safe_serialize(returns),
                        }
                    ),
                    result_datetime=get_current_ts(),
                    context_token=self.session_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
                if self.verbose:
                    print(
                        f"[Okareo] Logged function use '{thread_id}' for agent '{source_name}'"
                    )
            except Exception as e:
                print(f"[Okareo] Failed to log function use: {e}")
        else:
            pass

    def get_connection(self) -> None:
        """Method is intentionally left blank because there is no specific connection needed for the OkareoLogger."""

    def stop(self) -> None:
        """Close the file handler and remove it from the logger."""
        if self.verbose:
            print(f"[Okareo] Evaluating logging Session ID: {self.session_id}")
        self.okareo.create_trace_eval(self.group, self.session_id)
        print(
            f"[Okareo] Logged data points with:\n-> Context Token '{self.session_id}'\n-> Group Name '{self.group_name}'\n-> Group ID '{self.group['id']}'"
        )


class AutogenLogger:
    def __init__(
        self,
        config: dict[str, Any],
    ) -> None:
        self.logger = OkareoLogger(config)

    def __enter__(self) -> None:
        # start logging
        autogen.runtime_logging.start(logger=self.logger)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # stop logging
        autogen.runtime_logging.stop()



import json
import os
import time
from typing import List

import pytest
import requests  # type: ignore
from okareo_tests.common import API_KEY, random_string

from okareo import BaseGenerationSchema, Okareo
from okareo_api_client.models import (
    ScenarioSetCreate,
    ScenarioSetGenerate,
    ScenarioSetResponse,
    SeedData,
)
from okareo_api_client.models.scenario_data_poin_response import (
    ScenarioDataPoinResponse,
)
from okareo_api_client.models.scenario_type import ScenarioType


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


@pytest.fixture(scope="module")
def seed_data() -> List[SeedData]:
    return [
        SeedData(input_="example question or statement", result="example result"),
        SeedData(
            input_="tell me about your capability", result="unique example result"
        ),
        SeedData(input_="what are your limitations", result="different example result"),
    ]


@pytest.fixture(scope="module")
def synonym_data() -> List[List[str]]:
    return [
        ["sit", "sit amet"],
        ["elit", "bar"],
        ["labore", "foobar"],
    ]


@pytest.fixture(scope="module")
def custom_data() -> List[SeedData]:
    return [
        SeedData(input_="Lorem ipsum dolor sit amet", result="1"),
        SeedData(input_="consectetur adipiscing elit, sed do", result="2"),
        SeedData(input_="eiusmod tempor incididunt ut labore", result="3"),
    ]


create_scenario_name = f"my_test_scenario_set_{random_string(5)}"
synonym_scenario_name = f"my_synonym_scenario_set_{random_string(5)}"
examples_scenario_name = f"my_examples_scenario_set_{random_string(5)}"


@pytest.fixture(scope="module")
def create_scenario_set(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=create_scenario_name,
        seed_data=seed_data,
    )
    articles: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    return articles


@pytest.fixture(scope="module")
def create_examples_scenario_set(
    okareo_client: Okareo, custom_data: List[SeedData]
) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=examples_scenario_name,
        seed_data=custom_data,
    )
    examples: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    return examples


@pytest.fixture(scope="module")
def generate_scenarios(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_scenario_set.scenario_id,
        name=f"generated scenario set {random_string(5)}",
        number_examples=2,
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def generate_scenarios_qa(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_scenario_set.scenario_id,
        name=f"generated scenario set {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.TEXT_REVERSE_QUESTION_ANSWER,
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def generate_scenarios_synonym(
    okareo_client: Okareo,
    create_examples_scenario_set: ScenarioSetResponse,
    synonym_data: List[List[str]],
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_examples_scenario_set.scenario_id,
        name=f"generated scenario synonyms set {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.SYNONYMS,
        synonym_sets=synonym_data,
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def generate_scenarios_custom(
    okareo_client: Okareo, create_examples_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_examples_scenario_set.scenario_id,
        name=f"generated custom scenario set {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.CUSTOM_GENERATOR,
        generation_prompt="generate the next 5 words of 'lorem ipsum' based on the following text: {input}",
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def generate_scenarios_custom_multi_chunk(
    okareo_client: Okareo, create_examples_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_examples_scenario_set.scenario_id,
        name=f"generated custom multi-chunk scenario set {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.CUSTOM_MULTI_CHUNK_GENERATOR,
        generation_prompt="generate the next 5 words of 'lorem ipsum' based on the following text: {input}",
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


class InnerInput(BaseGenerationSchema):
    inner_foo: str
    inner_bar: int
    inner_baz: float


class ComplicatedJsonInput(BaseGenerationSchema):
    foo: str
    bar: int
    baz: float
    qux: dict
    tic: List[str]
    toc: List[InnerInput]


class GenerationSchema(BaseGenerationSchema):
    input: ComplicatedJsonInput
    result: str


@pytest.fixture(scope="module")
def generate_scenarios_custom_with_schema(
    okareo_client: Okareo, create_examples_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_examples_scenario_set.scenario_id,
        name=f"generated custom scenario set with schema {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.CUSTOM_GENERATOR,
        generation_prompt="generate the next 5 words of 'lorem ipsum' based on the following text: {input}",
        generation_schema=GenerationSchema,  # type: ignore
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def get_scenario_data_points(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> List[ScenarioDataPoinResponse]:
    assert isinstance(create_scenario_set.scenario_id, str)
    return okareo_client.get_scenario_data_points(create_scenario_set.scenario_id)


def assert_generated_scenarios_seed_ids(
    okareo_client: Okareo,
    generated_scenario: ScenarioSetResponse,
    expect_multiple_seeds: bool = False,
) -> None:
    """Assert that each generated scenario's seed_id exists in the source scenario's data points.

    Args:
        okareo_client: The Okareo client instance
        generated_scenario: The generated scenario response to validate
        expect_multiple_seeds: If True, expects seed_id to be a JSON string containing multiple IDs
    """
    gen_dp = okareo_client.get_scenario_data_points(generated_scenario.scenario_id)  # type: ignore
    seed_dp = okareo_client.get_scenario_data_points(
        generated_scenario.tags[0].split(":")[1]  # type: ignore
    )
    seed_ids = [dp.id for dp in seed_dp]

    for dp in gen_dp:
        assert dp.meta_data
        if expect_multiple_seeds:
            for gen_id in json.loads(dp.meta_data["seed_id"]):
                assert gen_id in seed_ids
        else:
            assert dp.meta_data["seed_id"] in seed_ids


def test_create_scenario_set(
    create_scenario_set: ScenarioSetResponse, seed_data: List[SeedData]
) -> None:
    assert create_scenario_set.type == "SEED"  # default test
    assert create_scenario_set.scenario_id
    assert create_scenario_set.project_id
    assert create_scenario_set.time_created
    if isinstance(create_scenario_set.seed_data, List):
        for i in range(1):
            assert seed_data[i].input_ == create_scenario_set.seed_data[i].input_
            assert seed_data[i].result == create_scenario_set.seed_data[i].result


def test_download_scenario_set(
    okareo_client: Okareo,
    create_scenario_set: ScenarioSetResponse,
    seed_data: List[SeedData],
) -> None:
    response_file = okareo_client.download_scenario_set(create_scenario_set)
    with open(response_file.name) as scenario_file:
        for line in scenario_file:
            assert json.loads(line)["input"] != ""
            assert json.loads(line)["result"] != ""

    os.remove(create_scenario_name + ".jsonl")


def test_duplicate_scenario_name(
    okareo_client: Okareo,
    seed_data: List[SeedData],
) -> None:
    name = f"my test scenario set {random_string(5)}"
    scenario = okareo_client.create_scenario_set(
        ScenarioSetCreate(
            name=name,
            seed_data=seed_data,
        )
    )
    assert scenario.warning is None
    scenario = okareo_client.create_scenario_set(
        ScenarioSetCreate(
            name=name,
            seed_data=seed_data,
        )
    )
    assert scenario.warning


def test_download_scenario_set_with_file(
    okareo_client: Okareo,
    create_scenario_set: ScenarioSetResponse,
    seed_data: List[SeedData],
) -> None:
    response_file = okareo_client.download_scenario_set(
        create_scenario_set, "./scenario.jsonl"
    )
    with open(response_file.name) as scenario_file:
        for line in scenario_file:
            assert json.loads(line)["input"] != ""
            assert json.loads(line)["result"] != ""

    os.remove("./scenario.jsonl")


def test_create_scenario_set_all_fields(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    scenario_set_create = ScenarioSetCreate(
        name=f"my test scenario set {random_string(5)}",
        seed_data=seed_data,
    )
    create_scenario_set: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    assert create_scenario_set.type == "SEED"
    assert create_scenario_set.scenario_id
    assert create_scenario_set.project_id
    assert create_scenario_set.time_created
    assert isinstance(create_scenario_set.scenario_input, List)
    assert isinstance(create_scenario_set.seed_data, List)
    for i in range(3):
        assert seed_data[i].input_ == create_scenario_set.seed_data[i].input_


def test_generate_scenarios(
    generate_scenarios: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios.type == "REPHRASE_INVARIANT"
    assert generate_scenarios.seed_data == []
    assert generate_scenarios is not None
    assert generate_scenarios.scenario_id
    assert generate_scenarios.project_id
    assert generate_scenarios.time_created
    assert type(generate_scenarios.tags) is list

    assert_generated_scenarios_seed_ids(okareo_client, generate_scenarios)


def test_generate_scenarios_qa(
    generate_scenarios_qa: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios_qa.type == "TEXT_REVERSE_QUESTION_ANSWER"
    assert generate_scenarios_qa.seed_data == []
    assert generate_scenarios_qa is not None
    assert generate_scenarios_qa.scenario_id
    assert generate_scenarios_qa.project_id
    assert generate_scenarios_qa.time_created
    assert type(generate_scenarios_qa.tags) is list

    assert_generated_scenarios_seed_ids(okareo_client, generate_scenarios_qa)


def test_generate_scenarios_synonyms(
    generate_scenarios_synonym: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios_synonym.type == "SYNONYMS"
    assert generate_scenarios_synonym.seed_data == []
    assert generate_scenarios_synonym is not None
    assert generate_scenarios_synonym.scenario_id
    assert generate_scenarios_synonym.project_id
    assert generate_scenarios_synonym.time_created
    assert type(generate_scenarios_synonym.tags) is list

    assert_generated_scenarios_seed_ids(okareo_client, generate_scenarios_synonym)


def test_generate_scenarios_custom(
    generate_scenarios_custom: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios_custom.type == "CUSTOM_GENERATOR"
    assert generate_scenarios_custom.seed_data == []
    assert generate_scenarios_custom is not None
    assert generate_scenarios_custom.scenario_id
    assert generate_scenarios_custom.project_id
    assert generate_scenarios_custom.time_created
    assert type(generate_scenarios_custom.tags) is list

    assert_generated_scenarios_seed_ids(okareo_client, generate_scenarios_custom)


def test_generate_scenarios_custom_multi_chunk(
    generate_scenarios_custom_multi_chunk: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios_custom_multi_chunk.type == "CUSTOM_MULTI_CHUNK_GENERATOR"
    assert generate_scenarios_custom_multi_chunk.seed_data == []
    assert generate_scenarios_custom_multi_chunk is not None
    assert generate_scenarios_custom_multi_chunk.scenario_id
    assert generate_scenarios_custom_multi_chunk.project_id
    assert generate_scenarios_custom_multi_chunk.time_created
    assert type(generate_scenarios_custom_multi_chunk.tags) is list

    assert_generated_scenarios_seed_ids(
        okareo_client, generate_scenarios_custom_multi_chunk
    )


def test_generate_scenarios_custom_with_schema(
    generate_scenarios_custom_with_schema: ScenarioSetResponse,
    seed_data: List[SeedData],
    okareo_client: Okareo,
) -> None:
    assert generate_scenarios_custom_with_schema.type == "CUSTOM_GENERATOR"
    assert generate_scenarios_custom_with_schema.seed_data == []
    assert generate_scenarios_custom_with_schema is not None
    assert generate_scenarios_custom_with_schema.scenario_id
    assert generate_scenarios_custom_with_schema.project_id
    assert generate_scenarios_custom_with_schema.time_created
    assert type(generate_scenarios_custom_with_schema.tags) is list

    assert_generated_scenarios_seed_ids(
        okareo_client, generate_scenarios_custom_with_schema
    )

    # assert the structure of the generated scenario rows matches the schema
    scenario_data = generate_scenarios_custom_with_schema.scenario_data
    assert scenario_data is not None
    for row in scenario_data:  # type: ignore
        assert row.input_ is not None
        assert isinstance(row.input_, dict)
        assert "foo" in row.input_ and isinstance(row.input_["foo"], str)
        assert "bar" in row.input_ and isinstance(row.input_["bar"], int)
        assert "baz" in row.input_ and isinstance(row.input_["baz"], float)
        assert "qux" in row.input_ and isinstance(row.input_["qux"], dict)
        assert "tic" in row.input_ and isinstance(row.input_["tic"], list)
        assert all(isinstance(item, str) for item in row.input_["tic"])
        assert "toc" in row.input_ and isinstance(row.input_["toc"], list)
        assert all(
            isinstance(item, dict)
            and "inner_foo" in item
            and "inner_bar" in item
            and "inner_baz" in item
            for item in row.input_["toc"]
        )
        assert row.result is not None
        assert isinstance(row.result, str)


def test_generate_scenarios_empty(
    okareo_client: Okareo,
    create_scenario_set: ScenarioSetResponse,
    synonym_data: List[List[str]],
) -> None:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_scenario_set.scenario_id,
        name=f"generated scenario synonyms set {random_string(5)}",
        number_examples=1,
        generation_type=ScenarioType.SYNONYMS,
        synonym_sets=synonym_data,
    )
    # expect a warning since no synonyms are found in the source scenario rows
    with pytest.warns(
        UserWarning,
        match="No scenario rows generated for scenario_id",
    ) as record:
        generated_scenario = okareo_client.generate_scenario_set(scenario_set_generate)
        assert generated_scenario is not None
        assert generated_scenario.scenario_id == ""
        assert generated_scenario.time_created
        assert generated_scenario.type == "SYNONYMS"
        if not record:
            pytest.fail("Expected warning for empty generated scenario.")


def test_get_scenario_data_points(
    get_scenario_data_points: List[ScenarioDataPoinResponse], seed_data: List[SeedData]
) -> None:
    assert get_scenario_data_points is not None
    assert len(get_scenario_data_points) == 3


def delete_scenario_data_points(
    api_key: str, scenario_id: str, scenario_name: str
) -> None:
    url = f"https://api.okareo.com/v0/scenario_sets/{scenario_id}?name={scenario_name}"
    headers = {
        "Content-Type": "application/json",
        "accept": "application/json",
        "api-key": api_key,
    }
    requests.delete(url, headers=headers)


def test_create_scenario_empty_seed_data(
    okareo_client: Okareo,
) -> None:
    scenario_set_create = ScenarioSetCreate(
        name=f"my test scenario set {random_string(5)}",
        seed_data=[],
    )
    with pytest.raises(
        ValueError, match="Non-empty seed data is required to create a scenario set"
    ):
        okareo_client.create_scenario_set(scenario_set_create)


def dtest_create_delete_scenario_set_contraction_tiny_load(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    seed_data = []
    file_path = os.path.join(
        os.path.dirname(__file__), "datasets/random_sentence_tiny.txt"
    )
    with open(file_path) as file:
        lines = file.readlines()
        for line in lines:
            parts = line.split("\t")
            assert len(parts) == 2
            seed_data.append(
                SeedData(input_=parts[0], result=parts[1]),
            )

    scenario_set_create = ScenarioSetCreate(
        name="my contraction test scenario set tiny load",
        seed_data=seed_data,
    )
    scenario: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    time.sleep(10)

    if isinstance(scenario.tags, list) and isinstance(scenario.name, str):
        scenario_seed_id = scenario.tags[0].split(":")[1]
        delete_scenario_data_points(API_KEY, scenario_seed_id, scenario.name)
        assert isinstance(scenario.scenario_id, str)
        delete_scenario_data_points(API_KEY, scenario.scenario_id, scenario.name)

    assert scenario.type == "COMMON_CONTRACTIONS"


def dtest_create_delete_scenario_set_misspelling_tiny_load(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    seed_data = []
    file_path = os.path.join(
        os.path.dirname(__file__), "datasets/random_sentence_tiny.txt"
    )
    with open(file_path) as file:
        lines = file.readlines()
        for line in lines:
            parts = line.split("\t")
            assert len(parts) == 2
            seed_data.append(
                SeedData(input_=parts[0], result=parts[1]),
            )

    scenario_set_create = ScenarioSetCreate(
        name=f"my misspelling test scenario set tiny load {random_string(5)}",
        seed_data=seed_data,
    )
    scenario: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    time.sleep(10)

    if isinstance(scenario.tags, list) and isinstance(scenario.name, str):
        scenario_seed_id = scenario.tags[0].split(":")[1]
        delete_scenario_data_points(API_KEY, scenario_seed_id, scenario.name)
        assert isinstance(scenario.scenario_id, str)

        delete_scenario_data_points(API_KEY, scenario.scenario_id, scenario.name)

    assert scenario.type == "COMMON_MISSPELLINGS"



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.filter_condition import FilterCondition


T = TypeVar("T", bound="DatapointFilterSearch")


@_attrs_define
class DatapointFilterSearch:
    """
    Attributes:
        filters (List['FilterCondition']): List of filter conditions to apply
        project_id (Union[Unset, str]): Project ID to search within
        offset (Union[Unset, int]): Offset for pagination
        limit (Union[Unset, int]): Limit for pagination
        issues_only (Union[Unset, bool]): Only return issues
        errors_only (Union[Unset, bool]): Only return errors
        checks (Union[Unset, List[Any]]): List of checks to only flag issues for
        filter_group_id (Union[Unset, str]): Filter group ID to search with
        timezone (Union[Unset, str]): IANA timezone to use for date filtering/aggregation (e.g., 'America/New_York').
            Defaults to None (i.e., UTC).
        precision (Union[Unset, str]): Time precision for the summary. Valid values include ['day', 'hour', 'minute'].
            Defaults to 'day'. Default: 'day'.
    """

    filters: List["FilterCondition"]
    project_id: Union[Unset, str] = UNSET
    offset: Union[Unset, int] = UNSET
    limit: Union[Unset, int] = UNSET
    issues_only: Union[Unset, bool] = False
    errors_only: Union[Unset, bool] = False
    checks: Union[Unset, List[Any]] = UNSET
    filter_group_id: Union[Unset, str] = UNSET
    timezone: Union[Unset, str] = UNSET
    precision: Union[Unset, str] = "day"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        filters = []
        for filters_item_data in self.filters:
            filters_item = filters_item_data.to_dict()

            filters.append(filters_item)

        project_id = self.project_id
        offset = self.offset
        limit = self.limit
        issues_only = self.issues_only
        errors_only = self.errors_only
        checks: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.checks, Unset):
            checks = self.checks

        filter_group_id = self.filter_group_id
        timezone = self.timezone
        precision = self.precision

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "filters": filters,
            }
        )
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if offset is not UNSET:
            field_dict["offset"] = offset
        if limit is not UNSET:
            field_dict["limit"] = limit
        if issues_only is not UNSET:
            field_dict["issues_only"] = issues_only
        if errors_only is not UNSET:
            field_dict["errors_only"] = errors_only
        if checks is not UNSET:
            field_dict["checks"] = checks
        if filter_group_id is not UNSET:
            field_dict["filter_group_id"] = filter_group_id
        if timezone is not UNSET:
            field_dict["timezone"] = timezone
        if precision is not UNSET:
            field_dict["precision"] = precision

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.filter_condition import FilterCondition

        d = src_dict.copy()
        filters = []
        _filters = d.pop("filters")
        for filters_item_data in _filters:
            filters_item = FilterCondition.from_dict(filters_item_data)

            filters.append(filters_item)

        project_id = d.pop("project_id", UNSET)

        offset = d.pop("offset", UNSET)

        limit = d.pop("limit", UNSET)

        issues_only = d.pop("issues_only", UNSET)

        errors_only = d.pop("errors_only", UNSET)

        checks = cast(List[Any], d.pop("checks", UNSET))

        filter_group_id = d.pop("filter_group_id", UNSET)

        timezone = d.pop("timezone", UNSET)

        precision = d.pop("precision", UNSET)

        datapoint_filter_search = cls(
            filters=filters,
            project_id=project_id,
            offset=offset,
            limit=limit,
            issues_only=issues_only,
            errors_only=errors_only,
            checks=checks,
            filter_group_id=filter_group_id,
            timezone=timezone,
            precision=precision,
        )

        datapoint_filter_search.additional_properties = d
        return datapoint_filter_search

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import random
import string
from datetime import datetime
from unittest import mock

from langchain.chains import LLMChain
from langchain.llms.fake import FakeListLLM
from langchain.prompts import PromptTemplate
from okareo_tests.common import API_KEY, OkareoAPIhost, integration, random_string
from pytest_httpx import HTTPXMock

from okareo import Okareo
from okareo.callbacks import CallbackHandler
from okareo_api_client.models.datapoint_search import DatapointSearch
from okareo_api_client.models.model_under_test_response import ModelUnderTestResponse

MOCK_GLOBAL_RESPONSE = [
    {
        "id": "0156f5d7-4ac4-4568-9d44-24750aa08d1a",
        "name": "Global",
        "onboarding_status": "onboarding_status",
        "tags": [],
        "additional_properties": {},
    }
]


def get_mut_response() -> dict:
    return ModelUnderTestResponse(
        "id",
        "my-project",
        "langchain_test",
        ["ci-run"],
        datetime.now().isoformat(),
    ).to_dict()


@integration
def test_llm_generates_datapoints(
    httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost
) -> None:
    context_token = random_string(10)
    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=MOCK_GLOBAL_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(json=get_mut_response(), status_code=201)

    with mock.patch.object(Okareo.__init__, "__defaults__", (okareo_api.path, 100)):
        handler = CallbackHandler(
            mut_name="langchain_test", context_token=context_token
        )

    llm = FakeListLLM(responses=["4"])
    llm.predict("what is 2+2?", callbacks=[handler])

    if okareo_api.is_mock:
        requests = httpx_mock.get_requests()
        print("requests", requests)
        assert requests[1].method == "POST"
        assert "/v0/register_model" in requests[1].url.path
        assert requests[2].method == "POST"
        assert "/v0/datapoints" in requests[2].url.path
    else:
        okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
        dp = okareo.find_datapoints(DatapointSearch(context_token=context_token))
        if isinstance(dp, list):
            assert len(dp) == 1


@integration
def test_llm_auto_generate_model(
    httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost
) -> None:
    context_token = "".join(random.choices(string.ascii_letters, k=10))
    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=MOCK_GLOBAL_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(json=get_mut_response(), status_code=201)

    with mock.patch.object(Okareo.__init__, "__defaults__", (okareo_api.path, 100)):
        handler = CallbackHandler(context_token=context_token)

    llm = FakeListLLM(responses=["4"])
    llm.predict("what is 2+2?", callbacks=[handler])

    if okareo_api.is_mock:
        requests = httpx_mock.get_requests()
        print("requests", requests)
        assert requests[1].method == "POST"
        assert "/v0/register_model" in requests[1].url.path
        assert requests[2].method == "POST"
        assert "/v0/datapoints" in requests[2].url.path
    else:
        okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
        dp = okareo.find_datapoints(DatapointSearch(context_token=context_token))
        if isinstance(dp, list):
            assert len(dp) == 1


@integration
def test_chain_generates_datapoints(
    httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost
) -> None:
    context_token = random_string(10)
    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=MOCK_GLOBAL_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(json=get_mut_response(), status_code=201)

    with mock.patch.object(Okareo.__init__, "__defaults__", (okareo_api.path, 100)):
        handler = CallbackHandler(context_token=context_token)

    llm = FakeListLLM(responses=["4"])
    prompt = PromptTemplate.from_template("what is 2+{number}?")
    chain = LLMChain(llm=llm, prompt=prompt, callbacks=[handler])
    chain.run(number=2)

    if okareo_api.is_mock:
        requests = httpx_mock.get_requests()
        print("requests", requests)
        assert requests[1].method == "POST"
        assert "/v0/register_model" in requests[1].url.path
        assert requests[2].method == "POST"
        assert "/v0/datapoints" in requests[2].url.path
    else:
        okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
        dp = okareo.find_datapoints(DatapointSearch(context_token=context_token))
        if isinstance(dp, list):
            assert len(dp) == 1



import os
from typing import Any, Union

import pytest
from okareo_tests.common import API_KEY, random_string
from okareo_tests.utils import assert_metrics

from okareo import Okareo
from okareo.checks import CheckOutputType, ModelBasedCheck
from okareo.model_under_test import (
    CustomModel,
    GenerationModel,
    ModelInvocation,
    MultiTurnDriver,
)
from okareo_api_client.models import ScenarioSetResponse
from okareo_api_client.models.comparison_operator import ComparisonOperator
from okareo_api_client.models.datapoint_field import DatapointField
from okareo_api_client.models.datapoint_filter_search import DatapointFilterSearch
from okareo_api_client.models.evaluator_spec_request import EvaluatorSpecRequest
from okareo_api_client.models.filter_condition import FilterCondition
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.seed_data import SeedData
from okareo_api_client.models.test_run_type import TestRunType

OPENAI_API_KEY = os.environ["OPENAI_API_KEY"]


@pytest.fixture(scope="module")
def rnd() -> str:
    return random_string(5)


@pytest.fixture(scope="module")
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


TEST_SUMMARIZE_TEMPLATE = """
Provide a brief summary of the following paragraph of text:

{scenario_input}

Summary:

"""


@pytest.fixture(scope="module")
def article_scenario_set(rnd: str, okareo: Okareo) -> ScenarioSetResponse:
    file_path = os.path.join(os.path.dirname(__file__), "webbizz_3_test_article.jsonl")
    articles: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=file_path, scenario_name=f"openai-scenario-set-{rnd}"
    )

    return articles


def test_generate_check(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    checks_to_generate = [
        {
            "description": "Return True if the model_output is at least 20 characters long, otherwise return False.",
            "requires_scenario_input": False,
            "requires_scenario_result": False,
            "name": "model_only",
        },
        {
            "description": "Return True if the scenario_result is at least 20 characters long, otherwise return False.",
            "requires_scenario_input": False,
            "requires_scenario_result": True,
            "name": "model_with_result",
        },
        {
            "description": "Return True if the scenario_input is at least 20 characters long, otherwise return False.",
            "requires_scenario_input": True,
            "requires_scenario_result": False,
            "name": "model_with_input",
        },
        {
            "description": "Return True if the combined length of the scenario_input and scenario_result is at least 20 characters long, otherwise return False.",
            "requires_scenario_input": True,
            "requires_scenario_result": True,
            "name": "model_with_input_and_result",
        },
    ]
    for check_dict in checks_to_generate:
        generate_request = EvaluatorSpecRequest(
            description=str(check_dict["description"]),
            requires_scenario_input=bool(check_dict["requires_scenario_input"]),
            requires_scenario_result=bool(check_dict["requires_scenario_result"]),
            output_data_type="bool",
        )
        check = okareo.generate_check(generate_request)

        assert check.generated_code


def test_run_code_based_predefined_checks(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    mut = okareo.register_model(
        name=f"openai-ci-run-levenshtein-{rnd}",
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )

    checks = [
        "levenshtein_distance",
        "levenshtein_distance_input",
        "compression_ratio",
        "does_code_compile",
        "contains_all_imports",
        "corpus_BLEU",
        "latency",
        "is_json",
    ]
    run_resp = mut.run_test(
        name=f"openai-chat-run-predefined-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=checks,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-predefined-{rnd}"
    assert_metrics(run_resp, checks, num_rows=3)


def test_run_model_based_predefined_checks(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    mut = okareo.register_model(
        name=f"openai-ci-run-levenshtein-{rnd}",
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )

    checks = ["consistency_summary", "fluency_summary", "relevance_summary"]
    run_resp = mut.run_test(
        name=f"openai-chat-run-predefined-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=checks,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-predefined-{rnd}"
    assert_metrics(run_resp, checks, num_rows=3)


def test_run_model_based_custom_checks(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    check_sample_pass_fail = okareo.create_or_update_check(
        name=f"check_sample_pass_fail {rnd}",
        description="check_sample_pass_fail",
        check=ModelBasedCheck(  # type: ignore
            prompt_template="Only output True if the model_output is at least 20 characters long, otherwise return False.",
            check_type=CheckOutputType.PASS_FAIL,
        ),
    )
    check_sample_score = okareo.create_or_update_check(
        name=f"check_sample_score {rnd}",
        description="check_sample_score",
        check=ModelBasedCheck(  # type: ignore
            prompt_template="Only output the number of words in the following text: {scenario_input} {output} {generation}",
            check_type=CheckOutputType.SCORE,
        ),
    )
    checks = [check_sample_score, check_sample_pass_fail]
    check_names = [check.name or "" for check in checks]
    mut = okareo.register_model(
        name=f"openai-ci-run-levenshtein-{rnd}",
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )
    run_resp = mut.run_test(
        name=f"openai-chat-run-predefined-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=check_names,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-predefined-{rnd}"
    assert_metrics(run_resp, check_names, num_rows=3)
    for check in checks:
        assert check.id
        assert check.name
        okareo.delete_check(check.id, check.name)


def test_run_code_based_custom_checks(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    from okareo_tests.checks.sample_check import Check  # type: ignore

    check_sample_code = okareo.create_or_update_check(
        name=f"check_sample_code {rnd}",
        description="check_sample_code",
        check=Check(),  # type: ignore
    )
    assert check_sample_code.id
    assert check_sample_code.name

    from okareo_tests.checks.sample_check_with_explanation import Check  # type: ignore

    check_sample_code_exp = okareo.create_or_update_check(
        name=f"check_sample_code_with_explanation {rnd}",
        description="check_sample_code_with_explanation",
        check=Check(),  # type: ignore
    )
    assert check_sample_code_exp.id
    assert check_sample_code_exp.name
    mut = okareo.register_model(
        name=f"openai-ci-run-levenshtein-{rnd}",
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )
    run_resp = mut.run_test(
        name=f"openai-chat-run-predefined-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=[check_sample_code.name, check_sample_code_exp.name],
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-predefined-{rnd}"
    assert_metrics(
        run_resp, [check_sample_code.name, check_sample_code_exp.name], num_rows=3
    )

    okareo.delete_check(check_sample_code.id, check_sample_code.name)
    okareo.delete_check(check_sample_code_exp.id, check_sample_code_exp.name)


@pytest.fixture(scope="module")
def large_scenario_set(rnd: str, okareo: Okareo) -> ScenarioSetResponse:
    """
    We are using 51 articles because that triggers parallel execution
    """
    file_path = os.path.join(
        os.path.dirname(__file__), "webbizz_51_test_articles.jsonl"
    )
    articles: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=file_path, scenario_name=f"webbizz-51-scenario-set-{rnd}"
    )

    return articles


def test_parallel_predefined_checks(
    rnd: str, okareo: Okareo, large_scenario_set: ScenarioSetResponse
) -> None:
    class FirstTenChars(CustomModel):
        def invoke(self, input_value: Union[dict, list, str]) -> Any:
            assert isinstance(input_value, str)

            return ModelInvocation(
                model_prediction=input_value[:10],
                model_input=input_value,
            )

    mut = okareo.register_model(
        name=f"custom-ci-run-10chars-{rnd}",
        model=FirstTenChars(name=f"custom-ci-run-10chars-{rnd}"),
    )

    checks = ["is_json"]
    run_resp = mut.run_test(
        name=f"custom-ci-run-parallel-predefined-{rnd}",
        scenario=large_scenario_set,
        test_run_type=TestRunType.NL_GENERATION,
        checks=checks,
    )
    assert run_resp.name == f"custom-ci-run-parallel-predefined-{rnd}"
    assert_metrics(run_resp, checks, num_rows=51)


def test_custom_check_on_multiturn_model_input_args(rnd: str, okareo: Okareo) -> None:
    # Ensure that the 'model_input' is always longer than the 'scenario_input'
    # The 'model_input' is the full list of messages
    # The 'scenario_input' is the most recent message to the target
    from okareo_tests.checks.input_comparison_check import Check  # type: ignore

    check_name = f"input_comparison_check_{rnd}"
    input_comparison_check = okareo.create_or_update_check(
        name=check_name,
        description="Check if model input is longer than scenario input",
        check=Check(),
    )

    scenario_set_create = ScenarioSetCreate(
        name=rnd,
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Will you help me with my homework?",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)

    mut = okareo.register_model(
        name=rnd + random_string(5),
        model=MultiTurnDriver(
            max_turns=1,
            repeats=1,
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
            stop_check={"check_name": "behavior_adherence", "stop_on": True},
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_name = "Model Input vs. Scenario Input Test"
    test_run_item = mut.run_test(
        scenario=response,
        api_key=OPENAI_API_KEY,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=[check_name],
    )
    assert test_run_item.name == test_run_name
    assert test_run_item.status == "FINISHED"

    assert test_run_item.model_metrics
    assert test_run_item.model_metrics.additional_properties
    metrics_dict = test_run_item.model_metrics.additional_properties
    assert metrics_dict.get("mean_scores") is not None
    assert metrics_dict["mean_scores"][check_name] == 1

    # cleanup: remove the check
    okareo.delete_check(input_comparison_check.id, input_comparison_check.name)  # type: ignore


def test_no_checks_on_every_turn(rnd: str, okareo: Okareo) -> None:
    scenario_set_create = ScenarioSetCreate(
        name=rnd,
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Will you help me with my homework?",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)

    mut = okareo.register_model(
        name=rnd + random_string(5),
        model=MultiTurnDriver(
            max_turns=2,
            repeats=1,
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_name = "No Checks on Every Turn"
    test_run_item = mut.run_test(
        scenario=response,
        api_key=OPENAI_API_KEY,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["behavior_adherence"],
    )
    assert test_run_item.name == test_run_name
    assert test_run_item.status == "FINISHED"

    assert test_run_item.model_metrics
    assert test_run_item.model_metrics.additional_properties
    metrics_dict = test_run_item.model_metrics.additional_properties
    assert metrics_dict.get("mean_scores") is not None
    assert metrics_dict["mean_scores"].get("behavior_adherence") is not None

    # Get the data_points where the test data point ID
    dp = okareo.find_datapoints_filter(
        DatapointFilterSearch(
            filters=[
                FilterCondition(
                    field=DatapointField.TEST_RUN_ID,
                    operator=ComparisonOperator.EQUAL,
                    value=test_run_item.id,
                )
            ]
        )
    )
    assert isinstance(dp, list)
    for i, data_point in enumerate(dp):
        assert data_point.test_data_point_id is not None
        assert data_point.test_run_id == test_run_item.id
        # dp are sorted by time_created (reverse chronological),
        # first dp should have checks
        # other dps should not have checks
        if i == 0:
            assert data_point.checks is not None
            assert len(data_point.checks.to_dict()) > 0  # type: ignore
            for check in data_point.checks.to_dict():  # type: ignore
                assert check in [
                    "behavior_adherence",
                    "behavior_adherence__explanation",
                    "avg_turn_latency",
                ]
        else:
            assert data_point.checks is None



import tempfile
from datetime import datetime
from typing import Any

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo_api_client.models import ScenarioType
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.seed_data import SeedData

# Create unique identifiers for test resources
today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")
rnd_str = random_string(5)
unique_key = f"{rnd_str} {today_with_time}"


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


class TestScenarioExceptions:
    """Tests for error paths related to scenario creation and manipulation."""

    def test_seed_data_missing(self, okareo_client: Okareo) -> None:
        """Test failure when seed data is missing"""
        scenario_set_create = ScenarioSetCreate(
            name=f"Test Missing Seed Data {unique_key}",
            seed_data=[],  # Empty seed data should cause an error
        )

        with pytest.raises(
            Exception, match="Non-empty seed data is required to create a scenario set"
        ):
            okareo_client.create_scenario_set(scenario_set_create)

    def test_name_validation_error(self, okareo_client: Okareo) -> None:
        """Test failure when scenario name is invalid"""
        scenario_set_create = ScenarioSetCreate(
            name="",  # Empty name should cause a validation error
            seed_data=[SeedData(input_="Test input", result="Test result")],
        )

        with pytest.raises(Exception, match="Scenario set name is invalid"):
            okareo_client.create_scenario_set(scenario_set_create)

    def test_duplicate_name(self, okareo_client: Okareo) -> None:
        """Test duplicate name handling"""
        unique_name = f"Test Duplicate Name {unique_key}"

        # First creation should succeed
        scenario_set_create = ScenarioSetCreate(
            name=unique_name,
            seed_data=[SeedData(input_="Test input", result="Test result")],
        )
        okareo_client.create_scenario_set(scenario_set_create)

        # Second creation with same name should return a warning
        # In this case, it shouldn't throw an exception but return the existing set
        response = okareo_client.create_scenario_set(scenario_set_create)
        assert (
            response.warning is not None
            and isinstance(response.warning, str)
            and "is already in use" in response.warning
        )

    def test_seed_data_processing_error(self, okareo_client: Okareo) -> None:
        """Test failure when there's an error processing seed data"""
        # Create problematic seed data structure
        # Using a nested structure that's too complex or has circular references
        complex_input: Any = {"data": {"nested": {"more_nested": {"circular": {}}}}}
        complex_input["data"]["nested"]["more_nested"][
            "circular"
        ] = complex_input  # Create circular reference

        try:
            # This should fail during JSON serialization
            scenario_set_create = ScenarioSetCreate(
                name=f"Test Processing Error {unique_key}",
                seed_data=[SeedData(input_=complex_input, result="Test result")],
            )
            okareo_client.create_scenario_set(scenario_set_create)
            pytest.fail("Expected exception was not raised")
        except Exception:
            # Either a client-side serialization error or server-side processing error
            assert True

    def test_file_validation_error_empty(self, okareo_client: Okareo) -> None:
        """Test error when uploading an empty file"""
        with tempfile.NamedTemporaryFile(suffix=".jsonl") as empty_file:
            # File is empty by default
            with pytest.raises(Exception, match="is empty"):
                okareo_client.upload_scenario_set(
                    file_path=empty_file.name,
                    scenario_name=f"Test Empty File {unique_key}",
                )

    def test_file_validation_error_invalid_json(self, okareo_client: Okareo) -> None:
        """Test error when uploading a file with invalid JSON"""
        with tempfile.NamedTemporaryFile(suffix=".jsonl", mode="w") as invalid_file:
            invalid_file.write('{"input": "test", "result": "test"}\n')  # Valid line
            invalid_file.write('{"input": "test", result: invalid}\n')  # Invalid JSON
            invalid_file.flush()

            with pytest.raises(Exception, match="Invalid JSON"):
                okareo_client.upload_scenario_set(
                    file_path=invalid_file.name,
                    scenario_name=f"Test Invalid JSON {unique_key}",
                )

    def test_file_validation_error_missing_fields(self, okareo_client: Okareo) -> None:
        """Test error when uploading a file with missing required fields"""
        with tempfile.NamedTemporaryFile(
            suffix=".jsonl", mode="w"
        ) as missing_fields_file:
            missing_fields_file.write(
                '{"only_input": "test"}\n'
            )  # Missing result field
            missing_fields_file.write(
                '{"only_result": "test"}\n'
            )  # Missing input field
            missing_fields_file.flush()

            with pytest.raises(Exception, match="must have 'input' and 'result'"):
                okareo_client.upload_scenario_set(
                    file_path=missing_fields_file.name,
                    scenario_name=f"Test Missing Fields {unique_key}",
                )

    def test_generation_with_invalid_source(self, okareo_client: Okareo) -> None:
        """Test scenario generation with invalid source scenario ID"""
        with pytest.raises(Exception, match="not found"):
            okareo_client.generate_scenarios(
                source_scenario="00000000-0000-0000-0000-000000000000",  # Non-existent scenario ID
                name=f"Test Invalid Source {unique_key}",
                number_examples=2,
                generation_type=ScenarioType.REPHRASE_INVARIANT,
            )



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="GeneralFindPayload")


@_attrs_define
class GeneralFindPayload:
    """
    Attributes:
        id (Union[Unset, str]): ID for the testrun
        project_id (Union[Unset, str]): ID for the project
        mut_id (Union[Unset, str]): ID of the model
        scenario_set_id (Union[Unset, str]): ID of the scenario set
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter test runs in the Okareo app
        return_model_metrics (Union[Unset, bool]): Boolean value indicating if model metrics should be returned. This
            increases the response size.
        return_error_matrix (Union[Unset, bool]): Boolean value indicating if error matrix should be returned. This
            increases the response size.
    """

    id: Union[Unset, str] = UNSET
    project_id: Union[Unset, str] = UNSET
    mut_id: Union[Unset, str] = UNSET
    scenario_set_id: Union[Unset, str] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    return_model_metrics: Union[Unset, bool] = False
    return_error_matrix: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        project_id = self.project_id
        mut_id = self.mut_id
        scenario_set_id = self.scenario_set_id
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        return_model_metrics = self.return_model_metrics
        return_error_matrix = self.return_error_matrix

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if scenario_set_id is not UNSET:
            field_dict["scenario_set_id"] = scenario_set_id
        if tags is not UNSET:
            field_dict["tags"] = tags
        if return_model_metrics is not UNSET:
            field_dict["return_model_metrics"] = return_model_metrics
        if return_error_matrix is not UNSET:
            field_dict["return_error_matrix"] = return_error_matrix

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id", UNSET)

        project_id = d.pop("project_id", UNSET)

        mut_id = d.pop("mut_id", UNSET)

        scenario_set_id = d.pop("scenario_set_id", UNSET)

        tags = cast(List[str], d.pop("tags", UNSET))

        return_model_metrics = d.pop("return_model_metrics", UNSET)

        return_error_matrix = d.pop("return_error_matrix", UNSET)

        general_find_payload = cls(
            id=id,
            project_id=project_id,
            mut_id=mut_id,
            scenario_set_id=scenario_set_id,
            tags=tags,
            return_model_metrics=return_model_metrics,
            return_error_matrix=return_error_matrix,
        )

        general_find_payload.additional_properties = d
        return general_find_payload

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from datetime import datetime
from typing import Union

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import CustomModel, ModelInvocation
from okareo_api_client.models import ScenarioSetResponse, TestRunType
from okareo_api_client.models.datapoint_search import DatapointSearch
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")
rnd_str = random_string(5)
unique_key = f"{rnd_str} {today_with_time}"
create_scenario_name = f"ci_json_test_create_scenarios {unique_key}"


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


JSON_SCENARIO = [
    {
        "input": {"query": "user q 1", "user_id": "red", "meta": "meta1"},
        "result": "red",
    },
    {
        "input": {"query": "user q 2", "user_id": "blue", "meta": "meta1"},
        "result": "blue",
    },
    {
        "input": {"query": "user q 3", "user_id": "green", "meta": "meta1"},
        "result": "green",
    },
]  # type: ignore
SCENARIO_USERID = [scenario["input"]["user_id"] for scenario in JSON_SCENARIO]  # type: ignore
SCENARIO_QUERY = [scenario["input"]["query"] for scenario in JSON_SCENARIO]  # type: ignore
SCENARIO_META = [scenario["input"]["meta"] for scenario in JSON_SCENARIO]  # type: ignore


JSON_SEED = Okareo.seed_data_from_list(JSON_SCENARIO)  # type: ignore


@pytest.fixture(scope="module")
def create_scenario_set(okareo_client: Okareo) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=create_scenario_name,
        seed_data=JSON_SEED,
    )
    scenario = okareo_client.create_scenario_set(scenario_set_create)
    return scenario


# BC test for custom model removed as it has been phased out of SDK and docs


def test_custom_return_invocation(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> None:
    test_run_name = f"ci_test_custom_return_bc {unique_key}"

    class ClassificationModel(CustomModel):
        def invoke(self, input_value: Union[dict, list, str]) -> ModelInvocation:
            assert isinstance(input_value, dict)

            return ModelInvocation(
                model_prediction=input_value["user_id"],
                model_input=input_value["query"],
                model_output_metadata=input_value["meta"],
            )

    model_under_test = okareo_client.register_model(
        name=test_run_name,
        model=ClassificationModel(name=test_run_name),
    )

    test_run = model_under_test.run_test(
        scenario=create_scenario_set,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics=True,
    )

    assert test_run.id
    assert test_run.model_metrics
    metrics_dict = test_run.model_metrics.to_dict()
    assert metrics_dict["weighted_average"] is not None
    assert metrics_dict["weighted_average"]["f1"] == 1

    dp = okareo_client.find_datapoints(DatapointSearch(test_run_id=test_run.id))
    assert isinstance(dp, list)
    assert len(dp) == 3
    for d in dp:
        assert d.input_ in SCENARIO_QUERY
        assert d.result in SCENARIO_USERID



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.datapoint_list_item_agent_metadata_type_0 import DatapointListItemAgentMetadataType0
    from ..models.datapoint_list_item_checks import DatapointListItemChecks
    from ..models.datapoint_list_item_input_tools_item import DatapointListItemInputToolsItem
    from ..models.datapoint_list_item_model_metadata_type_0 import DatapointListItemModelMetadataType0
    from ..models.datapoint_list_item_result_embeddings_item import DatapointListItemResultEmbeddingsItem
    from ..models.datapoint_list_item_result_tool_calls_item import DatapointListItemResultToolCallsItem
    from ..models.datapoint_list_item_user_metadata_type_0 import DatapointListItemUserMetadataType0


T = TypeVar("T", bound="DatapointListItem")


@_attrs_define
class DatapointListItem:
    """
    Attributes:
        id (str):
        tags (Union[Unset, List[str]]):
        input_ (Union[Unset, Any]):
        input_datetime (Union[Unset, datetime.datetime]):
        result (Union[Unset, Any]):
        result_datetime (Union[Unset, datetime.datetime]):
        feedback (Union[Unset, float]):
        error_message (Union[Unset, str]):
        error_code (Union[Unset, str]):
        error_type (Union[Unset, str]):
        context_token (Union[Unset, str]):
        time_created (Union[Unset, datetime.datetime]):
        model_metadata (Union['DatapointListItemModelMetadataType0', Any, Unset]):
        project_id (Union[Unset, str]):
        mut_id (Union[Unset, str]):
        test_run_id (Union[Unset, str]):
        test_data_point_id (Union[Unset, str]):
        group_id (Union[Unset, str]):
        scenario_data_point_id (Union[Unset, str]):
        latency (Union[Unset, int]):
        input_tokens (Union[Unset, int]):
        output_tokens (Union[Unset, int]):
        source (Union[Unset, str]):
        temperature (Union[Unset, str]):
        input_tools (Union[Unset, List['DatapointListItemInputToolsItem']]):
        result_tool_calls (Union[Unset, List['DatapointListItemResultToolCallsItem']]):
        result_embeddings (Union[Unset, List['DatapointListItemResultEmbeddingsItem']]):
        checks (Union[Unset, DatapointListItemChecks]):
        agent_metadata (Union['DatapointListItemAgentMetadataType0', Any, Unset]):
        provider (Union[Unset, str]):
        total_search_count (Union[Unset, int]):
        total_datapoint_count (Union[Unset, int]):
        request_model_name (Union[Unset, str]):
        response_model_name (Union[Unset, str]):
        cost (Union[Unset, float]):
        status (Union[Unset, str]):  Default: ''.
        failed_checks (Union[Unset, List[str]]): Array of failed check names
        resolved (Union[Unset, bool]): Manual marking from user on resolved status
        user_metadata (Union['DatapointListItemUserMetadataType0', Any, Unset]): User-provided metadata provided as
            context to the completion call.
    """

    id: str
    tags: Union[Unset, List[str]] = UNSET
    input_: Union[Unset, Any] = UNSET
    input_datetime: Union[Unset, datetime.datetime] = UNSET
    result: Union[Unset, Any] = UNSET
    result_datetime: Union[Unset, datetime.datetime] = UNSET
    feedback: Union[Unset, float] = UNSET
    error_message: Union[Unset, str] = UNSET
    error_code: Union[Unset, str] = UNSET
    error_type: Union[Unset, str] = UNSET
    context_token: Union[Unset, str] = UNSET
    time_created: Union[Unset, datetime.datetime] = UNSET
    model_metadata: Union["DatapointListItemModelMetadataType0", Any, Unset] = UNSET
    project_id: Union[Unset, str] = UNSET
    mut_id: Union[Unset, str] = UNSET
    test_run_id: Union[Unset, str] = UNSET
    test_data_point_id: Union[Unset, str] = UNSET
    group_id: Union[Unset, str] = UNSET
    scenario_data_point_id: Union[Unset, str] = UNSET
    latency: Union[Unset, int] = UNSET
    input_tokens: Union[Unset, int] = UNSET
    output_tokens: Union[Unset, int] = UNSET
    source: Union[Unset, str] = UNSET
    temperature: Union[Unset, str] = UNSET
    input_tools: Union[Unset, List["DatapointListItemInputToolsItem"]] = UNSET
    result_tool_calls: Union[Unset, List["DatapointListItemResultToolCallsItem"]] = UNSET
    result_embeddings: Union[Unset, List["DatapointListItemResultEmbeddingsItem"]] = UNSET
    checks: Union[Unset, "DatapointListItemChecks"] = UNSET
    agent_metadata: Union["DatapointListItemAgentMetadataType0", Any, Unset] = UNSET
    provider: Union[Unset, str] = UNSET
    total_search_count: Union[Unset, int] = UNSET
    total_datapoint_count: Union[Unset, int] = UNSET
    request_model_name: Union[Unset, str] = UNSET
    response_model_name: Union[Unset, str] = UNSET
    cost: Union[Unset, float] = UNSET
    status: Union[Unset, str] = ""
    failed_checks: Union[Unset, List[str]] = UNSET
    resolved: Union[Unset, bool] = False
    user_metadata: Union["DatapointListItemUserMetadataType0", Any, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.datapoint_list_item_agent_metadata_type_0 import DatapointListItemAgentMetadataType0
        from ..models.datapoint_list_item_model_metadata_type_0 import DatapointListItemModelMetadataType0
        from ..models.datapoint_list_item_user_metadata_type_0 import DatapointListItemUserMetadataType0

        id = self.id
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        input_ = self.input_
        input_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.input_datetime, Unset):
            input_datetime = self.input_datetime.isoformat()

        result = self.result
        result_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.result_datetime, Unset):
            result_datetime = self.result_datetime.isoformat()

        feedback = self.feedback
        error_message = self.error_message
        error_code = self.error_code
        error_type = self.error_type
        context_token = self.context_token
        time_created: Union[Unset, str] = UNSET
        if not isinstance(self.time_created, Unset):
            time_created = self.time_created.isoformat()

        model_metadata: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.model_metadata, Unset):
            model_metadata = UNSET

        elif isinstance(self.model_metadata, DatapointListItemModelMetadataType0):
            model_metadata = UNSET
            if not isinstance(self.model_metadata, Unset):
                model_metadata = self.model_metadata.to_dict()

        else:
            model_metadata = self.model_metadata

        project_id = self.project_id
        mut_id = self.mut_id
        test_run_id = self.test_run_id
        test_data_point_id = self.test_data_point_id
        group_id = self.group_id
        scenario_data_point_id = self.scenario_data_point_id
        latency = self.latency
        input_tokens = self.input_tokens
        output_tokens = self.output_tokens
        source = self.source
        temperature = self.temperature
        input_tools: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.input_tools, Unset):
            input_tools = []
            for input_tools_item_data in self.input_tools:
                input_tools_item = input_tools_item_data.to_dict()

                input_tools.append(input_tools_item)

        result_tool_calls: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.result_tool_calls, Unset):
            result_tool_calls = []
            for result_tool_calls_item_data in self.result_tool_calls:
                result_tool_calls_item = result_tool_calls_item_data.to_dict()

                result_tool_calls.append(result_tool_calls_item)

        result_embeddings: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.result_embeddings, Unset):
            result_embeddings = []
            for result_embeddings_item_data in self.result_embeddings:
                result_embeddings_item = result_embeddings_item_data.to_dict()

                result_embeddings.append(result_embeddings_item)

        checks: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.checks, Unset):
            checks = self.checks.to_dict()

        agent_metadata: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.agent_metadata, Unset):
            agent_metadata = UNSET

        elif isinstance(self.agent_metadata, DatapointListItemAgentMetadataType0):
            agent_metadata = UNSET
            if not isinstance(self.agent_metadata, Unset):
                agent_metadata = self.agent_metadata.to_dict()

        else:
            agent_metadata = self.agent_metadata

        provider = self.provider
        total_search_count = self.total_search_count
        total_datapoint_count = self.total_datapoint_count
        request_model_name = self.request_model_name
        response_model_name = self.response_model_name
        cost = self.cost
        status = self.status
        failed_checks: Union[Unset, List[str]] = UNSET
        if not isinstance(self.failed_checks, Unset):
            failed_checks = self.failed_checks

        resolved = self.resolved
        user_metadata: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.user_metadata, Unset):
            user_metadata = UNSET

        elif isinstance(self.user_metadata, DatapointListItemUserMetadataType0):
            user_metadata = UNSET
            if not isinstance(self.user_metadata, Unset):
                user_metadata = self.user_metadata.to_dict()

        else:
            user_metadata = self.user_metadata

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if tags is not UNSET:
            field_dict["tags"] = tags
        if input_ is not UNSET:
            field_dict["input"] = input_
        if input_datetime is not UNSET:
            field_dict["input_datetime"] = input_datetime
        if result is not UNSET:
            field_dict["result"] = result
        if result_datetime is not UNSET:
            field_dict["result_datetime"] = result_datetime
        if feedback is not UNSET:
            field_dict["feedback"] = feedback
        if error_message is not UNSET:
            field_dict["error_message"] = error_message
        if error_code is not UNSET:
            field_dict["error_code"] = error_code
        if error_type is not UNSET:
            field_dict["error_type"] = error_type
        if context_token is not UNSET:
            field_dict["context_token"] = context_token
        if time_created is not UNSET:
            field_dict["time_created"] = time_created
        if model_metadata is not UNSET:
            field_dict["model_metadata"] = model_metadata
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if test_run_id is not UNSET:
            field_dict["test_run_id"] = test_run_id
        if test_data_point_id is not UNSET:
            field_dict["test_data_point_id"] = test_data_point_id
        if group_id is not UNSET:
            field_dict["group_id"] = group_id
        if scenario_data_point_id is not UNSET:
            field_dict["scenario_data_point_id"] = scenario_data_point_id
        if latency is not UNSET:
            field_dict["latency"] = latency
        if input_tokens is not UNSET:
            field_dict["input_tokens"] = input_tokens
        if output_tokens is not UNSET:
            field_dict["output_tokens"] = output_tokens
        if source is not UNSET:
            field_dict["source"] = source
        if temperature is not UNSET:
            field_dict["temperature"] = temperature
        if input_tools is not UNSET:
            field_dict["input_tools"] = input_tools
        if result_tool_calls is not UNSET:
            field_dict["result_tool_calls"] = result_tool_calls
        if result_embeddings is not UNSET:
            field_dict["result_embeddings"] = result_embeddings
        if checks is not UNSET:
            field_dict["checks"] = checks
        if agent_metadata is not UNSET:
            field_dict["agent_metadata"] = agent_metadata
        if provider is not UNSET:
            field_dict["provider"] = provider
        if total_search_count is not UNSET:
            field_dict["total_search_count"] = total_search_count
        if total_datapoint_count is not UNSET:
            field_dict["total_datapoint_count"] = total_datapoint_count
        if request_model_name is not UNSET:
            field_dict["request_model_name"] = request_model_name
        if response_model_name is not UNSET:
            field_dict["response_model_name"] = response_model_name
        if cost is not UNSET:
            field_dict["cost"] = cost
        if status is not UNSET:
            field_dict["status"] = status
        if failed_checks is not UNSET:
            field_dict["failed_checks"] = failed_checks
        if resolved is not UNSET:
            field_dict["resolved"] = resolved
        if user_metadata is not UNSET:
            field_dict["user_metadata"] = user_metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.datapoint_list_item_agent_metadata_type_0 import DatapointListItemAgentMetadataType0
        from ..models.datapoint_list_item_checks import DatapointListItemChecks
        from ..models.datapoint_list_item_input_tools_item import DatapointListItemInputToolsItem
        from ..models.datapoint_list_item_model_metadata_type_0 import DatapointListItemModelMetadataType0
        from ..models.datapoint_list_item_result_embeddings_item import DatapointListItemResultEmbeddingsItem
        from ..models.datapoint_list_item_result_tool_calls_item import DatapointListItemResultToolCallsItem
        from ..models.datapoint_list_item_user_metadata_type_0 import DatapointListItemUserMetadataType0

        d = src_dict.copy()
        id = d.pop("id")

        tags = cast(List[str], d.pop("tags", UNSET))

        input_ = d.pop("input", UNSET)

        _input_datetime = d.pop("input_datetime", UNSET)
        input_datetime: Union[Unset, datetime.datetime]
        if isinstance(_input_datetime, Unset):
            input_datetime = UNSET
        else:
            input_datetime = isoparse(_input_datetime)

        result = d.pop("result", UNSET)

        _result_datetime = d.pop("result_datetime", UNSET)
        result_datetime: Union[Unset, datetime.datetime]
        if isinstance(_result_datetime, Unset):
            result_datetime = UNSET
        else:
            result_datetime = isoparse(_result_datetime)

        feedback = d.pop("feedback", UNSET)

        error_message = d.pop("error_message", UNSET)

        error_code = d.pop("error_code", UNSET)

        error_type = d.pop("error_type", UNSET)

        context_token = d.pop("context_token", UNSET)

        _time_created = d.pop("time_created", UNSET)
        time_created: Union[Unset, datetime.datetime]
        if isinstance(_time_created, Unset):
            time_created = UNSET
        else:
            time_created = isoparse(_time_created)

        def _parse_model_metadata(data: object) -> Union["DatapointListItemModelMetadataType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _model_metadata_type_0 = data
                model_metadata_type_0: Union[Unset, DatapointListItemModelMetadataType0]
                if isinstance(_model_metadata_type_0, Unset):
                    model_metadata_type_0 = UNSET
                else:
                    model_metadata_type_0 = DatapointListItemModelMetadataType0.from_dict(_model_metadata_type_0)

                return model_metadata_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointListItemModelMetadataType0", Any, Unset], data)

        model_metadata = _parse_model_metadata(d.pop("model_metadata", UNSET))

        project_id = d.pop("project_id", UNSET)

        mut_id = d.pop("mut_id", UNSET)

        test_run_id = d.pop("test_run_id", UNSET)

        test_data_point_id = d.pop("test_data_point_id", UNSET)

        group_id = d.pop("group_id", UNSET)

        scenario_data_point_id = d.pop("scenario_data_point_id", UNSET)

        latency = d.pop("latency", UNSET)

        input_tokens = d.pop("input_tokens", UNSET)

        output_tokens = d.pop("output_tokens", UNSET)

        source = d.pop("source", UNSET)

        temperature = d.pop("temperature", UNSET)

        input_tools = []
        _input_tools = d.pop("input_tools", UNSET)
        for input_tools_item_data in _input_tools or []:
            input_tools_item = DatapointListItemInputToolsItem.from_dict(input_tools_item_data)

            input_tools.append(input_tools_item)

        result_tool_calls = []
        _result_tool_calls = d.pop("result_tool_calls", UNSET)
        for result_tool_calls_item_data in _result_tool_calls or []:
            result_tool_calls_item = DatapointListItemResultToolCallsItem.from_dict(result_tool_calls_item_data)

            result_tool_calls.append(result_tool_calls_item)

        result_embeddings = []
        _result_embeddings = d.pop("result_embeddings", UNSET)
        for result_embeddings_item_data in _result_embeddings or []:
            result_embeddings_item = DatapointListItemResultEmbeddingsItem.from_dict(result_embeddings_item_data)

            result_embeddings.append(result_embeddings_item)

        _checks = d.pop("checks", UNSET)
        checks: Union[Unset, DatapointListItemChecks]
        if isinstance(_checks, Unset):
            checks = UNSET
        else:
            checks = DatapointListItemChecks.from_dict(_checks)

        def _parse_agent_metadata(data: object) -> Union["DatapointListItemAgentMetadataType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _agent_metadata_type_0 = data
                agent_metadata_type_0: Union[Unset, DatapointListItemAgentMetadataType0]
                if isinstance(_agent_metadata_type_0, Unset):
                    agent_metadata_type_0 = UNSET
                else:
                    agent_metadata_type_0 = DatapointListItemAgentMetadataType0.from_dict(_agent_metadata_type_0)

                return agent_metadata_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointListItemAgentMetadataType0", Any, Unset], data)

        agent_metadata = _parse_agent_metadata(d.pop("agent_metadata", UNSET))

        provider = d.pop("provider", UNSET)

        total_search_count = d.pop("total_search_count", UNSET)

        total_datapoint_count = d.pop("total_datapoint_count", UNSET)

        request_model_name = d.pop("request_model_name", UNSET)

        response_model_name = d.pop("response_model_name", UNSET)

        cost = d.pop("cost", UNSET)

        status = d.pop("status", UNSET)

        failed_checks = cast(List[str], d.pop("failed_checks", UNSET))

        resolved = d.pop("resolved", UNSET)

        def _parse_user_metadata(data: object) -> Union["DatapointListItemUserMetadataType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _user_metadata_type_0 = data
                user_metadata_type_0: Union[Unset, DatapointListItemUserMetadataType0]
                if isinstance(_user_metadata_type_0, Unset):
                    user_metadata_type_0 = UNSET
                else:
                    user_metadata_type_0 = DatapointListItemUserMetadataType0.from_dict(_user_metadata_type_0)

                return user_metadata_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointListItemUserMetadataType0", Any, Unset], data)

        user_metadata = _parse_user_metadata(d.pop("user_metadata", UNSET))

        datapoint_list_item = cls(
            id=id,
            tags=tags,
            input_=input_,
            input_datetime=input_datetime,
            result=result,
            result_datetime=result_datetime,
            feedback=feedback,
            error_message=error_message,
            error_code=error_code,
            error_type=error_type,
            context_token=context_token,
            time_created=time_created,
            model_metadata=model_metadata,
            project_id=project_id,
            mut_id=mut_id,
            test_run_id=test_run_id,
            test_data_point_id=test_data_point_id,
            group_id=group_id,
            scenario_data_point_id=scenario_data_point_id,
            latency=latency,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            source=source,
            temperature=temperature,
            input_tools=input_tools,
            result_tool_calls=result_tool_calls,
            result_embeddings=result_embeddings,
            checks=checks,
            agent_metadata=agent_metadata,
            provider=provider,
            total_search_count=total_search_count,
            total_datapoint_count=total_datapoint_count,
            request_model_name=request_model_name,
            response_model_name=response_model_name,
            cost=cost,
            status=status,
            failed_checks=failed_checks,
            resolved=resolved,
            user_metadata=user_metadata,
        )

        datapoint_list_item.additional_properties = d
        return datapoint_list_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.scenario_data_poin_response_meta_data import ScenarioDataPoinResponseMetaData


T = TypeVar("T", bound="ScenarioDataPoinResponse")


@_attrs_define
class ScenarioDataPoinResponse:
    """
    Attributes:
        id (str):
        input_ (Union[Unset, Any]):
        result (Union[Unset, Any]):
        meta_data (Union[Unset, ScenarioDataPoinResponseMetaData]):
    """

    id: str
    input_: Union[Unset, Any] = UNSET
    result: Union[Unset, Any] = UNSET
    meta_data: Union[Unset, "ScenarioDataPoinResponseMetaData"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        input_ = self.input_
        result = self.result
        meta_data: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.meta_data, Unset):
            meta_data = self.meta_data.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if input_ is not UNSET:
            field_dict["input"] = input_
        if result is not UNSET:
            field_dict["result"] = result
        if meta_data is not UNSET:
            field_dict["meta_data"] = meta_data

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.scenario_data_poin_response_meta_data import ScenarioDataPoinResponseMetaData

        d = src_dict.copy()
        id = d.pop("id")

        input_ = d.pop("input", UNSET)

        result = d.pop("result", UNSET)

        _meta_data = d.pop("meta_data", UNSET)
        meta_data: Union[Unset, ScenarioDataPoinResponseMetaData]
        if isinstance(_meta_data, Unset):
            meta_data = UNSET
        else:
            meta_data = ScenarioDataPoinResponseMetaData.from_dict(_meta_data)

        scenario_data_poin_response = cls(
            id=id,
            input_=input_,
            result=result,
            meta_data=meta_data,
        )

        scenario_data_poin_response.additional_properties = d
        return scenario_data_poin_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
